import numpy as np
class BaseEstimator():
    """
    For Ensuring format of Ensemble models
    id: for index,
    name: display name,
    class_def: model name,
    tune_grid: hyperparameters to tune,
    is_gpu_enabled: default: None
    preprocess_steps(in future): preprocess steps as per model
    """
    def __init__(
        self,id,name,class_def,tune_grid,args={},
        tune_args={},
        is_gpu_enabled=False,preprocess_steps={},random_state=45):
        
        if not args:
            args = {}
        if not tune_grid:
            tune_grid = {}
        if not tune_args:
            tune_args = {}
        if not preprocess_steps:
            preprocess_steps= {}
        self.args = args
        self.is_gpu_enabled = is_gpu_enabled
        self.tune_grid = tune_grid
        self.tune_args = tune_args
        self.preprocess_steps=preprocess_steps
        self.id= id
        self.name = name
        self.class_def = class_def


    def get_dict(self):
        """
        TO get model properties
        """
        d = [
            ("ID", self.id),
            ("Name", self.name),
                ("Class", self.class_def),
                ("Args", self.args),
                ("Tune Grid", self.tune_grid),
                ("Tune Args", self.tune_args),
        ]

        return dict(d)

# classification
class BaseAdaBoostClassifier(BaseEstimator):
    def __init__(self, ) :
        #logger = get_logger\(\)
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        from sklearn.ensemble import AdaBoostClassifier

        args = {"random_state": 45}
        tune_args = {}
        tune_grid = {
            "n_estimators": np.arange(10, 300, 10, ),
            "learning_rate": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "algorithm": ["SAMME", "SAMME.R"],
        }
        super().__init__(
            id="ada",
            name="Ada Boost Classifier",
            class_def=AdaBoostClassifier(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
        )


class BaseGradientBoostingClassifier(BaseEstimator):
    def __init__(self, ) :
        #logger = get_logger\(\)
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        from sklearn.ensemble import GradientBoostingClassifier

        args = {"random_state": 45}
        tune_args = {}
        tune_grid = {
            "n_estimators": np.arange(10, 300, 10, ),
            "learning_rate": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "subsample": np.arange(0.2, 1, 0.05, ),
            "min_samples_split": [2, 4, 5, 7, 9, 10],
            "min_samples_leaf": [1, 2, 3, 4, 5],
            "max_depth": np.arange(1, 11, 1, ),
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "max_features": [ "sqrt", "log2"],
        }
        super().__init__(
            id="gbc",
            name="Gradient Boosting Classifier",
            class_def=GradientBoostingClassifier(),
            args=args,
            tune_grid=tune_grid,
            
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
        )

class BaseExtraTreesClassifier(BaseEstimator):
    pass

class BaseXGBClassifier(BaseEstimator):
    def __init__(self, ) :
        # logger = get_logger()
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        from xgboost import XGBClassifier

        args = {
        }
        tune_args = {}
        tune_grid = {
            "learning_rate": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "n_estimators": np.arange(10, 300, 10, ),
            "subsample": [0.2, 0.3, 0.5, 0.7, 0.9, 1],
            "max_depth": np.arange(1, 11, 1, ),
            "colsample_bytree": [0.5, 0.7, 0.9, 1],
            "min_child_weight": [1, 2, 3, 4],
            "reg_alpha": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                1,
                2,
                3,
                4,
                5,
                10,
            ],
            "reg_lambda": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                1,
                2,
                3,
                4,
                5,
                10,
            ],
            "scale_pos_weight": np.arange(0.5, 50, 0.1, ),
            "use_label_encoder":[False],
            "eval_metric":['merror'],
        }
        super().__init__(
            id="xgboost",
            name="Extreme Gradient Boosting",
            class_def=XGBClassifier(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
        )

class BaseCatBoostClassifier(BaseEstimator):
    pass

class BaseBaggingClassifier(BaseEstimator):
    def __init__(self, ) :
        #logger = get_logger\(\)
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        from sklearn.ensemble import BaggingClassifier

        args = {
            "random_state": 45,
            "n_jobs": 1,
        }
        tune_args = {}
        tune_grid = {
            "bootstrap": [True, False],
            "bootstrap_features": [True, False],
            "max_samples": np.arange(0.5, 1, 0.2),
        }

        super().__init__(
            id="Bagging",
            name="Bagging Classifier",
            class_def=BaggingClassifier(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
            
        )


class BaseStackingClassifier(BaseEstimator):
    def __init__(self, ) :
        #logger = get_logger\(\)
        
        from sklearn.ensemble import StackingClassifier
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        args = {}
        tune_args = {}
        tune_grid = {}
        

        

        super().__init__(
            id="Stacking",
            name="Stacking Classifier",
            class_def=StackingClassifier(),
            args=args,
            tune_grid=tune_grid,
            
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
            
        )


class BaseVotingClassifier(BaseEstimator):
    def __init__(self, ) :
        from sklearn.ensemble import VotingClassifier

        args = {}
        tune_args = {}
        tune_grid = {}

        super().__init__(
            id="Voting",
            name="Voting Classifier",
            class_def=VotingClassifier(estimators=[]),
            args=args,
            tune_grid=tune_grid,
            
            tune_args=tune_args,
            
        )


# regression
class BaseExtraTreesRegressor(BaseEstimator):
    pass

class BaseAdaBoostRegressor(BaseEstimator):
    def __init__(self, ) -> None:
        
        from sklearn.ensemble import AdaBoostRegressor

        args = {"random_state": 45}
        tune_args = {}
        tune_grid = {
            "n_estimators": np.arange(10, 300, 10),
            "learning_rate": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "loss": ["linear", "square", "exponential"],
        }
        preprocess_steps={'scale_data':True,'dummify_categoricals':True}
        super().__init__(
            id="ada",
            name="AdaBoost Regressor",
            class_def=AdaBoostRegressor(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps = preprocess_steps,
        )


class BaseGradientBoostingRegressor(BaseEstimator):
    def __init__(self, ) -> None:
        

        from sklearn.ensemble import GradientBoostingRegressor

        args = {"random_state": 45}
        tune_args = {}
        tune_grid = {
            "n_estimators": np.arange(10, 300, 10),
            "learning_rate": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "subsample": np.arange(0.2, 1, 0.05),
            "min_samples_split": [2, 4, 5, 7, 9, 10],
            "min_samples_leaf": [1, 2, 3, 4, 5],
            "max_depth": np.arange(1, 11, 1),
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "max_features": ["sqrt", "log2"],
        }
        preprocess_steps={'scale_data':True,'dummify_categoricals':True}
        super().__init__(
            id="gbr",
            name="Gradient Boosting Regressor",
            class_def=GradientBoostingRegressor(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps = preprocess_steps,
        )


class BaseXGBRegressor(BaseEstimator):
    def __init__(self, ) -> None:
        
        try:
            import xgboost
        except ImportError:
            print("Couldn't import xgboost.XGBRegressor")
            
            return
        from xgboost import XGBRegressor

        args = {
        }
        tune_args = {}
        tune_grid = {
            "booster": ["gbtree"],
            "tree_method":  ["auto"],
            "learning_rate": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "n_estimators": np.arange(10, 300, 10),
            "subsample": [0.2,  1, 0.3, 0.5,0.7, 0.9],
            "max_depth": np.arange(1, 11, 1),
            "colsample_bytree": [0.5, 0.7, 1, 0.9],
            "min_child_weight": [1,  4,2, 3],
            "reg_alpha": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                1,
                2,
                3,
                4,
                5,
                10,
            ],
            "reg_lambda": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
                0.7,
                1,
                2,
                3,
                4,
                5,
                10,
            ],
            "scale_pos_weight": np.arange(1, 50, 0.1),
        }
        preprocess_steps={'scale_data':False,'dummify_categoricals':True}
        super().__init__(
            id="xgboost",
            name="Extreme Gradient Boosting",
            class_def=XGBRegressor(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
        )
  
class BaggingRegressor(BaseEstimator):
    def __init__(self, ) -> None:
        
        from sklearn.ensemble import BaggingRegressor

        args = {
            "random_state": 45,
            "n_jobs": -1,
        }
        tune_args = {}
        tune_grid = {
            "bootstrap": [True, False],
            "bootstrap_features": [True, False],
            "max_samples": np.arange(0.4, 1, 0.1),
        }
        super().__init__(
            id="Bagging",
            name="Bagging Regressor",
            class_def=BaggingRegressor(),
            args=args,
            tune_grid=tune_grid,
            
            tune_args=tune_args,
        )


class BaseStackingRegressor(BaseEstimator):
    def __init__(self, ) -> None:
        
        from sklearn.ensemble import StackingRegressor

        args = {}
        tune_args = {}
        tune_grid = {}       

        super().__init__(
            id="Stacking",
            name="Stacking Regressor",
            class_def=StackingRegressor(),
            args=args,
            tune_grid=tune_grid,
            
            tune_args=tune_args,
        )

class BaseVotingRegressor(BaseEstimator):
    def __init__(self, ) :
        from sklearn.ensemble import VotingRegressor

        args = {}
        tune_args = {}
        tune_grid = {}
        estimators = []

        super().__init__(
            id="Voting",
            name="Voting Regressor",
            class_def=VotingRegressor(estimators=estimators),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            
        )




from sklearn.base import BaseEstimator
from automl.models.utils import *
from sklearn.base import clone
# import matplotlib.pyplot as plt
import scikitplot as skplt

class ClassifyModel(BaseEstimator):
    '''
    Create and returns model for classification with best parameters
    using bayesian optimisation.
    Pass id as estimator (refer utils.py, classify_model) to create standalone model
    eg:
    >>> model = Classify_model(X,y,estimator='dt')
    >>> model.fit()    
    >>> model.predict(X)
    '''
    def __init__(self,X,y,estimator):
        self.modelbase = classify_models[estimator]
        self.modelname=self.modelbase.name
        self.x_train,self.y_train = X,y
        self.model = clone(self.modelbase.class_def)

    def fit(self,X=None,y=None):
        if (X==None) and (y==None):
            X = self.x_train
            y = self.y_train
        tune_grid = self.modelbase.tune_grid
        
        self.best_param,self.best_cv_score = get_best_param(estimator=self.model,X = X,y = y,tune_grid=tune_grid,)
        self.model.set_params(**self.best_param)
        self.model.fit(X,y)
        
    def get_params(self,deep=True): 
        return (self.best_param)
    
    def best_score(self):
        return self.best_cv_score
        
    def predict(self,X,y=None):
        predictions = self.model.predict(X)
        return predictions
    
    def predict_proba(self,X,y=None):
        predictions = self.model.predict_proba(X)
        return predictions
    
    def score(self,X,y):
        return self.model.score(X,y)
    
    def plot_model(self):
        probas = self.model.predict_proba(self.x_test)
        # Now plot.

        skplt.metrics.plot_precision_recall( self.y_test,probas)
        img = plot_config(title=self.modelname,xlabel='FPR',ylabel='TPR')
        return img



import numpy as np

class BaseRegressor():
    """For Ensuring format of regression model
    id: for index,
    name: display name,
    class_def: model name,
    tune_grid: hyperparameters to tune,
    is_gpu_enabled: default: None
    preprocess_steps(in future): preprocess steps as per model
    """
    def __init__(
        self,id,name,class_def,tune_grid,args={},
        tune_args={},
        is_gpu_enabled=False,preprocess_steps={},random_state=45):
        
        if not args:
            args = {}
        if not tune_grid:
            tune_grid = {}
        if not tune_args:
            tune_args = {}
        if not preprocess_steps:
            preprocess_steps= {}
        self.args = args
        self.is_gpu_enabled = is_gpu_enabled
        self.tune_grid = tune_grid
        self.tune_args = tune_args
        self.preprocess_steps={}

        self.is_boosting_supported = True
        self.is_soft_voting_supported = True
        self.id= id
        self.name = name
        self.class_def = class_def


    def get_dict(self):
        """
        TO get model properties
        """
        d = [
            ("ID", self.id),
            ("Name", self.name),
                ("Class", self.class_def),
                ("Args", self.args),
                ("Tune Grid", self.tune_grid),
                ("Tune Args", self.tune_args),
        ]

        return dict(d)

class BaseLinearRegressor(BaseRegressor):

    def __init__(self) -> None:
        from sklearn.linear_model import LinearRegression
        args = {}
        tune_args = {}
        tune_grid = {"fit_intercept": [True, False], "normalize": [True, False]} #
        # preprocess_steps = {'scale_data':True,'dummify_categoricals':True}

        super().__init__(
            id="lr",
            name="Linear Regression",
            class_def=LinearRegression(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            # preprocess_steps = preprocess_steps,
        )

class BaseLassoRegressor(BaseRegressor):

    def __init__(self) -> None:
        from sklearn.linear_model import Lasso
        args = {}
        tune_args = {}
        tune_grid = {
            "alpha": np.arange(0.01, 10, 0.01),
            "fit_intercept": [True, False],
            "normalize": [True, False],
        }
        preprocess_steps = {'scale_data':True,'dummify_categoricals':True}

        super().__init__(
            id="lasso",
            name="Lasso Regression",
            class_def=Lasso(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps = preprocess_steps,
        )

class BaseRidgeRegressor(BaseRegressor):

    def __init__(self) -> None:
        from sklearn.linear_model import Ridge
        args = {}
        tune_args = {}
        tune_grid = {
            "alpha": np.arange(0.01, 10, 0.01),
            "fit_intercept": [True, False],
            "normalize": [True, False],
        }
        preprocess_steps = {'scale_data':True,'dummify_categoricals':True}
        super().__init__(
            id="elnet",
            name="Elastic Regression",
            class_def=Ridge(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps = preprocess_steps,
        )


class BaseElasticNetRegressor(BaseRegressor):

    def __init__(self) -> None:
        from sklearn.linear_model import ElasticNet
        args = {}
        tune_args = {}
        tune_grid = {
            "alpha": np.arange(0.01, 10, 0.01),
            "l1_ratio": np.arange(0.01, 1, 0.001),
            "fit_intercept": [True, False],
            "normalize": [True, False],
        }
        preprocess_steps = {'scale_data':True,'dummify_categoricals':True}
        super().__init__(
            id="elnet",
            name="Elastic Regression",
            class_def=ElasticNet(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps = preprocess_steps,
        )

class BaseSVRRegressor(BaseRegressor):
    def __init__(self, ) -> None:
        
        from sklearn.svm import SVR
        args = {}
        tune_args = {}
        tune_grid = {
            "C": np.arange(0, 10, 0.001),
            "epsilon": [1.1, 1.2, 1.3, 1.35, 1.4, 1.5, 1.55, 1.6, 1.7, 1.8, 1.9],
        }
        preprocess_steps = {'scale_data':True,'dummify_categoricals':True}
        super().__init__(
            id="svr",
            name="Support Vector Regression",
            class_def=SVR(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps = preprocess_steps,
        )

class BaseKNeighborsRegressor(BaseRegressor):
    def __init__(self, ) -> None:
    
        from sklearn.neighbors import KNeighborsRegressor
        args = {}
        tune_args = {}
        tune_grid = {"n_neighbors":np.arange(2,10),
                     "weights":["uniform"],
                     "metric":["minkowski", "euclidean", "manhattan"]}
        preprocess_steps = {'scale_data':True,'dummify_categoricals':True}


        super().__init__(
            id="knn",
            name="K Neighbors Regressor",
            class_def=KNeighborsRegressor(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps = preprocess_steps,
        )

class BaseDecisionRegressor(BaseRegressor):
    def __init__(self,) -> None:
        from sklearn.tree import DecisionTreeRegressor

        args = {}
        tune_args = {}
        tune_grid = {
            "max_depth": np.arange(1, 16, 1),
            "max_features": [ "sqrt", "log2"],
            "min_samples_leaf": [2, 3, 4, 5, 6],
            "min_samples_split": [2, 5, 7, 9, 10],
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "criterion": ["mse", "mae", "friedman_mse"],
        }
        preprocess_steps={'scale_data':False,'dummify_categoricals':True}

        super().__init__(
            id="dt",
            name="Decision Tree Regressor",
            class_def=DecisionTreeRegressor(),
            args=args,
            tune_grid=tune_grid,
            
            tune_args=tune_args,
            preprocess_steps = preprocess_steps,
        )

class BaseRandomForRegressor(BaseRegressor):
    def __init__(self, ) -> None:
        

        from sklearn.ensemble import RandomForestRegressor
      
        tune_args = {}
        tune_grid = {
            "n_estimators": np.arange(10, 300, 10),
            "max_depth": np.arange(1, 11, 1),
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "max_features": [ "sqrt", "log2"],
            "bootstrap": [True, False],
        }
        preprocess_steps={'scale_data':False,'dummify_categoricals':True}
        super().__init__(
            id="rf",
            name="Random Forest Regressor",
            class_def=RandomForestRegressor(),
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
        )



import numpy as np
from skopt.space import Real, Categorical, Integer

class BaseClassifier():
    """For Ensuring format of classification  model
    id: for index,
    name: display name,
    class_def: model name,
    tune_grid: hyperparameters to tune,
    is_gpu_enabled: default: None
    preprocess_steps(in future): preprocess steps as per model
    """
    def __init__(
        self,id,name,class_def,tune_grid,args={},
        tune_args={},
        is_gpu_enabled=False,preprocess_steps={},random_state=45,space=[]):
        
        if not args:
            args = {}
        if not tune_grid:
            tune_grid = {}
        if not tune_args:
            tune_args = {}
        if not preprocess_steps:
            preprocess_steps= {}
        self.args = args
        self.tune_grid = tune_grid
        self.tune_args = tune_args
        self.preprocess_steps=preprocess_steps
        self.id= id
        self.name = name
        self.class_def = class_def
        self.space = space


    def get_dict(self):
        """
        TO get model properties
        """
        d = [
            ("ID", self.id),
            ("Name", self.name),
                ("Class", self.class_def),
                ("Args", self.args),
                ("Tune Grid", self.tune_grid),
                ("Tune Args", self.tune_args),
        ]

        return dict(d)

#noq
class BaseLogisticRegression(BaseClassifier):
    def __init__(self, ) :
        
        from sklearn.linear_model import LogisticRegression

        
        args = {}
        tune_args = {}
        tune_grid = {"C":np.arange(0.001, 10, 0.001,),
        "max_iter": [1000],
        "solver":["saga","lbfgs"],
        "penalty":["l2"],
        "class_weight" :["balanced"]
        }
        
        
        preprocess_steps = {'scale_data':True,'dummify_categoricals':True,'fit_le':True}
        super().__init__(
            id="lr",
            name="Logistic Regression",
            class_def=LogisticRegression(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
        )

#noq
class BaseKNeighborsClassifier(BaseClassifier):
    def __init__(self, ) :
        from sklearn.neighbors import KNeighborsClassifier
        args = {}
        tune_args = {}
        tune_grid = {"n_neighbors":np.arange(1, 51),"weights":["uniform"],"metric": ["minkowski", "euclidean", "manhattan"]}
        
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        super().__init__(
            id="knn",
            name="K Neighbors Classifier",
            class_def=KNeighborsClassifier(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
        )


class BaseGaussianNB(BaseClassifier):
    def __init__(self, ) :
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        from sklearn.naive_bayes import GaussianNB
        args = {}
        tune_args = {}
        space = [Real(0.0000001,0.1,name='var_smoothing')]
        tune_grid = {
            "var_smoothing": [
                0.000000001,
                0.00000005,
                0.0000008,
                0.000009,
                0.00001,
                0.001,
                0.007,
                0.009,
                0.01,
                0.1,
                1,
            ],
            "priors":[None],
        }
        
        super().__init__(
            id="nb",
            name="Naive Bayes",
            class_def=GaussianNB(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            space=space,
            preprocess_steps=preprocess_steps,
        )

#noq
class BaseDecisionTreeClassifier(BaseClassifier):
    def __init__(self, ):
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        from sklearn.tree import DecisionTreeClassifier

        args = {}
        tune_args = {}
        tune_grid = {
            "max_depth": np.arange(1, 16, 1, ),
            "max_features": [ "sqrt", "log2"],
            "min_samples_leaf": [2, 3, 4, 5, 6],
            "min_samples_split": [2, 5, 7, 9, 10],
            "criterion": ["gini", "entropy"],
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
        }
        super().__init__(
            id="dt",
            name="Decision Tree Classifier",
            class_def=DecisionTreeClassifier(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps=preprocess_steps
        )

#noq
class BaseSGDClassifier(BaseClassifier):
    def __init__(self, ) :
        preprocess_steps = {'scale_data':True,'dummify_categoricals':False}
        from sklearn.linear_model import SGDClassifier


        args = {}
        tune_args = {}
        tune_grid = {
            "penalty": ["elasticnet", "l2", "l1"],
            "l1_ratio": np.arange(0.0000000001, 1, 0.01),
            "alpha": [
                0.0000001,
                0.000001,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.15,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "fit_intercept": [True, False],
            "learning_rate": ["constant", "invscaling", "adaptive", "optimal"],
            "eta0": [0.001, 0.01, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5],
        }
        super().__init__(
            id="sgd",
            name="SVM - Linear Kernel",
            class_def=SGDClassifier(),
            args=args,
            tune_grid=tune_grid,
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
        )


class BaseSVC(BaseClassifier):
    def __init__(self, ) :
        preprocess_steps = {'scale_data':True,'dummify_categoricals':False}
        from sklearn.svm import SVC   
        args = {
            
        }
        tune_args = {}
        tune_grid = {
            "gamma": ["auto"],
            "probability": [True],
            "kernel": ["rbf"],
            "random_state": [45],
            "C": np.arange(0, 50, 0.01), #>0
            "class_weight": ["balanced"],
        }
        super().__init__(
            id="svc",
            name="SVM - Radial Kernel",
            class_def=SVC(),
            args=args,
            tune_grid=tune_grid,
            
            tune_args=tune_args,
            preprocess_steps=preprocess_steps,
        )


class BaseRandomForestClassifier(BaseClassifier):
    def __init__(self, ) :
        preprocess_steps = {'scale_data':False,'dummify_categoricals':False}
        from sklearn.ensemble import RandomForestClassifier

        tune_args = {}
        tune_grid = {
            "n_estimators": np.arange(10, 300, 10, ),
            "max_depth": np.arange(1, 11, 1, ),
            "min_impurity_decrease": [
                0,
                0.0001,
                0.001,
                0.01,
                0.0002,
                0.002,
                0.02,
                0.0005,
                0.005,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
            ],
            "max_features": ["sqrt", "log2"],
            "bootstrap": [True, False],
        }
        

        super().__init__(
            id="rf",
            name="Random Forest Classifier",
            class_def=RandomForestClassifier(),
            tune_grid=tune_grid,
            
            tune_args=tune_args,
        )





from sklearn.base import BaseEstimator
from automl.models.utils import *
from sklearn.base import clone
import logging
logging.basicConfig(level=logging.INFO)

class ClassifBaggingEnsemble(BaseEstimator):
    ''' Supported: Voting Classifier,
        default: Voting Classifier
        eg:
        >>> modelbase = BaggingEnsemble(X,y)
        >>> modelbase.create_bagging_model() #adds base models to bag
        >>> modelbase.fit()
        >>> model.predict(X)

    '''
   
    def __init__(self,X,y,estimator):
        self.estimator = estimator
        self.x_train,self.y_train = X,y
        self.modelbase = classif_bag_model[estimator]
        self.model = clone(self.modelbase.class_def)
        self.modelname = self.modelbase.name

    def base_model_grids(self,id):
        # changing grid of every base estimator as per required by esemble model for hyper parameter tuning
        modelbase = classify_models[id]
        modelname=modelbase.name
        model_class = modelbase.class_def
        tune_grid = modelbase.tune_grid
        
        tune_grid = {f'{id}__{k}':v for k,v in tune_grid.items()}
        
        return tune_grid
    
    def base_model(self,id):
        '''Creating base models of Bag'''
        modelbase = classify_models[id]
        modelname=modelbase.name
        model_class = modelbase.class_def
        model = clone(model_class)
        return model
    
    def fit(self,X=None,y=None):
        if (X==None) and (y==None):
            X = self.x_train
            y = self.y_train        
        self.model.set_params(estimators=self.base_ensemble_models())
        tune_grid = self.modelbase.tune_grid
        self.best_params,self.best_cv_score = get_best_param(estimator=self.model,X =X,y = y,tune_grid=tune_grid)
        self.model.set_params(**self.best_params)
        self.model.fit(self.x_train,self.y_train)
        # return model
    
    def get_params(self):
        return self.best_params

    def best_score(self):
        return self.best_cv_score
    
    def create_bagging_model(self):
        '''supported: voting classifier,
           ensemble_id:'vc' for voting classifier,
            base_models: rf,dt,lr (classification)
           '''
        tune_grid = {}
        estimators  = self.base_ensemble_models()
        for i in estimators:
            for k,v in self.base_model_grids(i[0]).items():
                tune_grid[k] = v
        self.modelbase.tune_grid = tune_grid

    def predict(self,X,y=None):
        return self.model.predict(X)
    
    def base_ensemble_models(self):
        # Instantiate classifiers
        svc = self.base_model(id='svc')
        dt = self.base_model(id = 'dt')
        nb = self.base_model(id='nb')
        estimators = [('svc',svc),('dt',dt),('nb',nb)]
        return estimators
 
    def score(self,X,y):
        return self.model.score(X,y)

class RegressBaggingEnsemble(BaseEstimator):
    ''' Supported: Voting Regressor,
        default: Voting Regressor
        eg:
        >>> modelbase = BaggingEnsemble(X,y)
        >>> modelbase.create_bagging_model() #adds base models to bag
        >>> modelbase.fit()
        >>> model.predict(X)

    '''
   
    def __init__(self,X,y,estimator):
        self.estimator = estimator
        self.x_train,self.y_train = X,y
        self.modelbase = regress_bag_model[estimator]
        self.model = clone(self.modelbase.class_def)
        self.modelname = self.modelbase.name
    
    def base_model_grids(self,id):
        # changing grid of every base estimator as per required by esemble model for hyper parameter tuning
        modelbase = regress_models[id]
        modelname=modelbase.name
        model_class = modelbase.class_def
        tune_grid = modelbase.tune_grid
        
        tune_grid = {f'{id}__{k}':v for k,v in tune_grid.items()}
        
        return tune_grid

    def base_model(self,id):
        '''Creating base models of Bag'''
        modelbase = regress_models[id]
        modelname=modelbase.name
        model_class = modelbase.class_def
        model = clone(model_class)
        return model
    
    def fit(self,X=None,y=None):
        if (X==None) and (y==None):
            X = self.x_train
            y = self.y_train        
        self.model.set_params(estimators=self.base_ensemble_models())
        tune_grid = self.modelbase.tune_grid
        self.best_params,self.best_cv_score = get_best_param(estimator=self.model,X = self.x_train,y = self.y_train,tune_grid=tune_grid)
        self.model.set_params(**self.best_params)
        self.model.fit(self.x_train,self.y_train)
        # return model

    def get_params(self):
        return self.best_params

    def best_score(self):
        return self.best_cv_score
        
    def create_bagging_model(self):
        '''supported: voting regressor,
           ensemble_id:'vc' for voting regressor,
            base_models: lr,svr (regression)
           '''
        tune_grid = {}
        estimators  = self.base_ensemble_models()
        for i in estimators:
            for k,v in self.base_model_grids(i[0]).items():
                tune_grid[k] = v
        self.modelbase.tune_grid = tune_grid

    def predict(self,X,y=None):
        return self.model.predict(X)

    def base_ensemble_models(self):
        # Instantiate Regressions
        lasso = self.base_model(id='lasso')
        dt = self.base_model(id='dt') 
        estimators = [('lasso',lasso),('dt',dt)]
        return estimators
    
    def score(self,X,y):
        return self.model.score(X,y)
 
#since not many changes in classification vs regression fitting technique they are combined        
class BoostingEnsemble(BaseEstimator):
    ''' Supported Xgboost
          eg:
        >>> modelbase = BoostingEnsemble(data=data,target=target,ml_type='regression')
        >>> model = modelbase.create_boosting_model()
        >>> model.fit()
        >>> modelbase.score() # accuracy of model based on 2% hold out train data set
        >>> model.predict(X)

    '''
    def __init__(self,X,y,ml_type,estimator='xgboost'):
        self.ml_type = ml_type
        self.x_train, self.y_train = X,y
        if self.ml_type=='regression':
            self.modelbase = regress_boost_model[estimator]
        else:
            self.modelbase = classif_boost_model[estimator]
        self.modelname = self.modelbase.name
        self.model = clone(self.modelbase.class_def)

    def fit(self,X=None,y=None):
        if (X==None):
            X = self.x_train   
        if (y==None):
            y = self.y_train
        tune_grid=self.modelbase.tune_grid
        self.best_params,self.best_cv_score = get_best_param(estimator=self.model,X = self.x_train,y = self.y_train,tune_grid=tune_grid,)
        self.model.set_params(**self.best_params)
        self.model.fit(self.x_train,self.y_train)

    def get_params(self,deep=True):
        return (self.best_params)

    def best_score(self):
        return self.best_cv_score
    def predict(self,X,y=None):
        predictions = self.model.predict(X)
        return predictions
    
    def score(self,X,y):
        test_score = self.model.score(X,y)
        return test_score



import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import clone
from automl.models.utils import *

class RegressModel(BaseEstimator):
    '''
    Create and returns model for regression with best parameters
    using bayesian optimisation.
    Pass estimator id (refer utils.py, regress_model) create standalone model
        eg:
    >>> modelbase = RegressModel(X,y,estimator='lr')
    >>> model.fit()
    >>> model.predict(X)
    >>> model.score(X,y)

    '''
    def __init__(self,X,y,estimator):

        self.modelbase = regress_models[estimator]
        self.modelname=self.modelbase.name
        self.x_train, self.y_train = X,y
        self.model = clone(self.modelbase.class_def)

    def fit(self,X=None,y=None):
        if (X==None) and (y==None):
            X = self.x_train
            y = self.y_train
        tune_grid = self.modelbase.tune_grid
        self.best_param,self.best_cv_score = get_best_param(estimator=self.model,X = X,y = y,tune_grid=tune_grid,)
        self.model.set_params(**self.best_param)
        self.model.fit(X,y)

    def get_params(self,deep=True):
        return (self.best_param)

    def best_score(self):
        return self.best_cv_score
    def predict(self,X,y=None):
        predictions = self.model.predict(X)
        return predictions

    def score(self,X,y):
        return self.model.score(X,y)

    def plot_model(self):
        train_acc = self.score(self.x_train, self.y_train)
        test_acc = self.score(self.x_train, self.y_train)
        plt.bar(x=['train accuracy','test accuracy'],height=[train_acc,test_acc],width=0.4)
        img = plot_config(title='r2',xlabel='data',ylabel='accuracy')
        return img
        



from automl.models.basemodels.classifiers import *
from automl.models.basemodels.regressors import *
from automl.models.basemodels.ensemblers import *
from skopt import BayesSearchCV
import matplotlib.pyplot as plt
from warnings import simplefilter
from sklearn.exceptions import ConvergenceWarning
simplefilter("ignore", category=ConvergenceWarning)

plt.switch_backend('agg')
import base64
from io import BytesIO
import logging
logging.basicConfig(level=logging.INFO)
classify_models={
    'lr':BaseLogisticRegression(),
         'dt':BaseDecisionTreeClassifier(),
         'svc':BaseSVC(),
         'rf':BaseRandomForestClassifier(),
         'knn':BaseKNeighborsClassifier(),
     'sgd':BaseSGDClassifier(),
     'nb':BaseGaussianNB(),
}

regress_models={
    'lr':BaseLinearRegressor(),
     'lasso':BaseLassoRegressor(),
     'dt':BaseDecisionRegressor(),
     'elnet':BaseElasticNetRegressor(),
         'knn':BaseKNeighborsRegressor(),
         'svr':BaseSVRRegressor(),
         'rf':BaseRandomForRegressor(),
         'knn':BaseKNeighborsRegressor(),
}

classif_bag_model = {
    'vc':BaseVotingClassifier(),
}
classif_boost_model ={
    'xgboost':BaseXGBClassifier()
}
regress_bag_model = {
    'vc':BaseVotingRegressor(),
}
regress_boost_model={
    'xgboost':BaseXGBRegressor()
}
def get_best_param(estimator,tune_grid,X,y):
    logging.info(f"Tune grid: {tune_grid}")
    logging.info(f"Model name: {estimator}")
    grid = BayesSearchCV(estimator=estimator,search_spaces=tune_grid,cv=10,n_jobs=-1,verbose=0) #cv=kfold default
    
    logging.info(f'Total iterations: {grid.total_iterations}')
    grid.fit(X,y,callback=on_step)
    logging.info(f"Best score:  {grid.best_score_}")
    logging.info(f"Best Parameters: {grid.best_params_}")
    return (grid.best_params_,grid.best_score_)
    
def on_step(optim_result):
    score = -optim_result['fun']
    logging.info(f"best score: %s" % score)
    if score >= 0.90:
        logging.info(f'Interrupting!')
        return True

def plot_config(**kwargs):
        # plt.figure(figsize=kwargs.get('figsize',(8,8)))
        plt.title(kwargs.get('title',''))
        plt.xlabel(kwargs.get("xlabel",'feature'))
        plt.ylabel(kwargs.get('ylabel','stat'))
        tmpfile = BytesIO()
        plt.savefig(tmpfile, format='png')
        plt.close()
        encoded = base64.b64encode(tmpfile.getvalue()).decode('utf-8')    
        
        return encoded

# from skopt.utils import use_named_args

# @use_named_args(space)
# def objective(**params):
#     reg.set_params(**params)

#     return -np.mean(cross_val_score(reg, X, y, cv=5, n_jobs=-1,
#                                     scoring="neg_mean_absolute_error"))

# def best_space(estimator,tune_grid,X,y):
    


from logging import log
from automl.models.utils import *
from automl.models.mlalgos import regression,ensemble
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_absolute_error,mean_squared_error,r2_score
from automl.preprocessing.preprocessor import preprocess_path
import inspect

logging.basicConfig(level=logging.INFO)
class BestRegessionModel():
    '''
    0) perform preprocessing
        Default:
        1) Remove duplicate rows
        2) Remove duplicate columns
        3) Remove columns with zero standard deviation
        4) Remove datetime columns
        5) perform standard scaler
        6) dummifiy categoricals

    1) split data into train_ and test_ data

    2) get best standalone regression model out of n (train_)
        Repeat for every regression model (specified in utils.py)
        1) split data in random k-folds(10)
        2) perform hyper parameter tuning with bayesian optimisation
        3) select the model with best cv score

    3) get best boosting model out of n (train_)
        Repeat for every boosting model (specified in utils.py)
        1) split data in random k-folds(10)
        2) perform hyper parameter tuning with bayesian optimisation
        3) select the model with best cv score
    
    4) get best bagging model out of n (train_)
        Repeat for every bagging model (specified in utils.py)
        1) select all base models (specified in utils.py)
        2) select tune grid of all models and add to bagging model's tune grid
        3) perform hyper parameter tuning with bayesian optimisation
              (basically same as optimizing base model parameters)
        4) select the model with best cv score
    
    5) select best model (from standalone,bag,boost) with rmse score (test_)

    6) Again perform fit to meta data with parameter tuning

    Eg:
    >>>    data_train = pd.read_csv('file')
    >>>    target = 'variety'
    >>>    model = BestRegessionModel(data = data_train,target=target)
    >>>    model.fit()
    
    >>>    data_test = pd.read_csv('file')
    >>>    predictions = model.predict(data_test)   
    >>>    print(f1score(test_y,predictions))   
    '''
    def __init__(self,data,target,preprocess=True,preprocess_steps={},*args,**kwargs):
        self.data = data
        self.target = target
        self.scores_grid = {}
        self.preprocess = preprocess
        if preprocess:
            if preprocess_steps:
                steps = inspect.getargspec(preprocess_path).args
                for key in preprocess_steps.keys():
                    if key not in steps:
                        raise ValueError("Invalid Parameter")
                try:
                    self.preprocess_pipe = preprocess_path(data=self.data,ml_type='regression',target=target,**preprocess_steps)
                    print("Done")
                    print(self.preprocess_pipe)
                except Exception as e:
                    logging.info(f"Invalid value passed {e}")
                    raise ValueError("Invalid value passed to parameter")
            else:
                self.preprocess_pipe = preprocess_path(data=self.data,ml_type='regression',scale_data=True,scaling_method="zscore",
                    target=self.target,dummify_categoricals=True,cluster_entire_data=False)
            
    def fit(self):
        logging.info(f'Started ,FGetting Best model')
        if self.preprocess:
            logging.info(f"Preprocessing the Training data")  
            self.data = self.preprocess_pipe.fit_transform(self.data)
        X=self.data.drop(self.target,axis=1)
        if 'cluster_label' in X.columns:
            X.drop('cluster_label',axis=1,inplace=True)
        y = self.data[self.target]
        self.meta_x_train,self.meta_x_test,self.meta_y_train,self.meta_y_test = train_test_split(X,y,test_size=0.33)
        self.standalone_model_ = self.get_best_standalone_estimate(self.meta_x_train,self.meta_y_train)
        self.boost_model_ = self.get_best_boost(self.meta_x_train,self.meta_y_train)
        self.bag_model_ = self.get_best_bag(self.meta_x_train,self.meta_y_train)
        standalone_ = self.standalone_model_.predict(self.meta_x_test)
        boost_ = self.boost_model_.predict(self.meta_x_test)
        bag_ = self.bag_model_.predict(self.meta_x_test)

        self.scores_grid = {
            'standalone': {
                'r2': r2_score(self.meta_y_test,standalone_),
                'mae':mean_absolute_error(self.meta_y_test,standalone_),
                'mse': mean_squared_error(self.meta_y_test,standalone_),
                'rmse': mean_squared_error(self.meta_y_test,standalone_,squared=False),
            },
            'boost': {
                'r2': r2_score(self.meta_y_test,boost_),
                'mae':mean_absolute_error(self.meta_y_test,boost_),
                'mse': mean_squared_error(self.meta_y_test,boost_),
                'rmse': mean_squared_error(self.meta_y_test,boost_,squared=False),
            },
            'bag': {
                'r2': r2_score(self.meta_y_test,bag_),
                'mae':mean_absolute_error(self.meta_y_test,bag_),
                'mse': mean_squared_error(self.meta_y_test,bag_),
                'rmse': mean_squared_error(self.meta_y_test,bag_,squared=False),
            },

        }
        logging.info(f'Model Scores:{self.scores_grid}')
        max_score = float('-inf')
        for model,method in self.scores_grid.items():
            if method['rmse'] >= max_score:
                max_score = method['rmse']
                if model == 'standalone':
                    max_model =self.standalone_model_
                elif model == 'boost':
                    max_model =self.boost_model_
                elif model == 'bag':
                    max_model = self.bag_model_
        logging.info(f'Best model was,:{max_model}')
        self.max_model = max_model
        self.max_model.model.fit(X,y)

    def get_best_standalone_estimate(self,x_train,y_train):
        logging.info(f'Looking for best standalone model')
        # regress_best_model = {} for storing complete info of best model
        self.all_regress_model_score = {}
        score_ =float('-inf')
        estimates = ['lr', 'dt','lasso','svr','knn','elnet','rf']

        for i in estimates:
            # model.drop(['cluster_label'],axis=1,inplace=True)
            model = regression.RegressModel(X=x_train,y=y_train,estimator=i)
            logging.info(f'Fitting standalone model: {i}')
            model.fit()
            logging.info(f'standalone model complete: {i}')
            self.all_regress_model_score[i] = {
                'modelname':model.modelname,
                'modelscore':model.best_score(),
                'modelparams':model.get_params(),
            }
            if model.best_score() > score_:
                score_ = model.best_cv_score # update score
                best_model = model
        logging.info("Best model Found, {best_model")
        return best_model

    def get_best_boost(self,x_train,y_train):
        logging.info(f'Looking for best boosting model')
        self.all_boost_model_score={}
        estimates = ['xgboost']
        score_ =float('-inf')
        for i in estimates:
            logging.info(f'Fitting Boosting model: {i}')
            model = ensemble.BoostingEnsemble(X=x_train,y=y_train,estimator=i,ml_type='regression')
            model.fit()
            logging.info(f'Fitting Complete: {i}')
            self.all_boost_model_score[i] = {
                'modelname':model.modelname,
                'modelscore':model.best_score(),
                'modelparams':model.get_params(),
            }

            if model.best_score() > score_:
                score_ = model.best_cv_score # update score
                best_model = model
        logging.info(f'Best model finding Complete:{best_model}')
        return best_model

    def get_best_bag(self,x_train,y_train):
        logging.info(f'Looking for best bagging model')
        all_bag_model_score={}
        estimates = ['vc']
        score_ =float('-inf')
        for i in estimates:
            logging.info(f'Fitting Bagging model: {i}')
            model = ensemble.RegressBaggingEnsemble(X=x_train,y=y_train,estimator=i)
            model.create_bagging_model()
            logging.info(f'Bagg created model: {i}')
            model.fit()
            logging.info(f'Fitting Bag  model: {i}')
            all_bag_model_score[i] = {
                'modelname':model.modelname,
                'modelscore':model.best_score(),
                'modelparams':model.get_params(),
            }
            
            if model.best_score() > score_:
                score_ = model.best_cv_score # update score
                best_model = model
        logging.info(f'Found model: {best_model}')
        return best_model

    def predict(self,X):
        X =self.preprocess_pipe.transform(X)
        if 'cluster_label' in X.columns:
            X.drop('cluster_label',axis=1,inplace=True)
        return self.max_model.model.predict(X)

    
    # def preprocess_data(self,**kwargs):
    #     logging.info(f'Started Preprocessing to train data')
    #     self.preprocess_pipe = preprocess_path(data=self.data,ml_type='regression',scale_data=True,scaling_method="zscore",
    #                     target=self.target,dummify_categoricals=True,cluster_entire_data=False,**kwargs)
    #     self.preprocess_pipe.fit_transform(self.data)

    # def inverse_preprocess_data(self,data):
    #     logging.info(f'Started Preprocessing for prediction data')
    #     self.preprocess_pipe.transform(data)
    #     return data

    # def get_preprocess_pipe(self):
    #     return self.preprocess_pipe


from setuptools import setup, find_packages

setup(name='light-automl',
    version='1.0.1',
    packages=find_packages(where="."),
    author='Atufa Shireen',
    package_dir={"": "."},
    url = "https://github.com/AtufaShireen/lightautoml",
    
    classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
  install_requires=[
    'joblib==1.1.0',
    'kneed==0.7.0',
    'matplotlib==3.5.0',
    'numpy==1.21.4',
    'pandas==1.3.4',  
    'scikit-learn==0.24.2',
    'scikit-optimize==0.9.0',
    'scikit-plot==0.3.7',
    'scipy==1.7.3',
    'xgboost==1.5.0',

  ],)




from automl.models import getregression,getclassification
from joblib import load,dump
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score,precision_score,recall_score,accuracy_score
from sklearn.metrics import mean_absolute_error,mean_squared_error
##-------------------------Classification Problem---------------------------#
df = pd.read_csv('diabetes_1.csv',nrows=1000)
# print(df.head())
target = 'Outcome'
x_train,x_test,y_train,y_test = train_test_split(df.drop(target,axis=1),df[target],test_size=0.2,stratify=df[target],random_state=45)
train_df = pd.concat([x_train,y_train],axis=1)
test_df = pd.concat([x_test,y_test],axis=1)

##---------------------------Training-------------------------------#
basemodel = getclassification.BestClassificationModel(train_df,target)
basemodel.fit() #fits to best model
dump(basemodel.preprocess_pipe,'classif_preprocess_pipe.joblib')
metrics = basemodel.scores_grid
print(metrics)
dump(basemodel,'best-classif-model.joblib')
dump(metrics,'classif-metrics.joblib')

# #---------------------------Testing---------------------------------#
pipe = load('classif_preprocess_pipe.joblib')
model = load('best-classif-model.joblib')
print(model.max_model.modelname)
print(model.max_model.get_params())
predictions = model.predict(x_test)
predictions = model.get_inverse_label(predictions)
# #-------------------------Score-------------------------------------#
n_df = pd.DataFrame(list(zip(predictions,y_test)),columns=['predicts','actual'])
n_df['predicts'] = n_df['predicts'].apply(lambda x: 1 if x=='Yes' else 0)
n_df['actual'] = n_df['actual'].apply(lambda x: 1 if x=='Yes' else 0)
print(n_df.head())
# print('CORRECT:',model.score(n_df['actual'],n_df['predicts'],normalize=False),"TOTAL:",len(n_df['actual']))
print('SCORE:',accuracy_score(n_df['actual'],n_df['predicts']))


##------------------------------Regression Problem--------------------#

df = pd.read_csv('house_price.csv',nrows=50)
print(df.shape[0])
target = 'TARGET(PRICE_IN_LACS)'
x_train,x_test,y_train,y_test = train_test_split(df.drop(target,axis=1),df[target],test_size=0.2,random_state=45)
train_df = pd.concat([x_train,y_train],axis=1)
test_df = pd.concat([x_test,y_test],axis=1)

# #---------------------------Training-------------------------------#
basemodel = getregression.BestRegessionModel(train_df,target)
basemodel.fit() #fits to best model
dump(basemodel.preprocess_pipe,'regress_preprocess_pipe.joblib')
metrics = basemodel.scores_grid
dump(basemodel,'best-regress-model.joblib')
dump(metrics,'regress-metrics.joblib')
print(metrics)
##---------------------------Testing---------------------------------#
# pipey = load('preprocess_pipe.joblib')
model = load('best-regress-model.joblib')
predictions = basemodel.predict(x_train)
##-------------------------Score-------------------------------------#
n_df = pd.DataFrame(list(zip(predictions,y_train)),columns=['predicts','actual'])
print(n_df.head())
print('SCORE:',mean_squared_error(n_df['actual'],n_df['predicts'],squared=False))
print(basemodel.scores_grid)



#@title preprocess, remove (remove duplicate cols,remove zero variance cols)
from sklearn.base import BaseEstimator 
import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder,OneHotEncoder
# from utils import _reverse_label_encoder
from sklearn.impute import SimpleImputer,KNNImputer
from sklearn.preprocessing import StandardScaler,MinMaxScaler,PowerTransformer,QuantileTransformer
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score,calinski_harabasz_score
from sklearn.pipeline import Pipeline
# from imblearn.over_sampling import SMOTE
# from imblearn.under_sampling import RandomUnderSampler
from kneed import KneeLocator
SKLEARN_EMPTY_STEP = "passthrough"
import logging

logging.basicConfig(level=logging.INFO)
# class InferDtypes():
#     # convdrting to float
    # def __init__(self,target) -> None:
        # self.target = target
    # def fit(self,dataset,y=None):
    #     data = dataset
    #     for i in data.select_dtypes(include=["object"]).columns:
    #         try:
    #             data[i] = data[i].astype("float32")
    #         except:
    #             None
#         #have str col names
#         data.columns = [str(i) for i in data.columns]
#         # remove special chars from col names
#         data.columns = data.columns.str.replace(r"[\,\}\{\]\[\:\"\']", "")
#         # if data type is bool or pandas Categorical , convert to categorical
#         # for i in data.select_dtypes(include=["bool", "category"]).columns:
#         #     data[i] = data[i].astype("object") 
#          # if int or float col has less than 5 unique vals convert to object: (exclude target)
#         for i in data.select_dtypes(include=["number"]).columns:
#             # if i != self.target:
#             if data[i].nunique() <= 5:  # hard coded
#                 data[i] = data[i].apply(str_if_not_null)
#             else:
#                 data[i] = data[i].astype("float32")
        # data.drop_duplicates(inplace=True)
        #         #infer datetime type
        #         for i in (
        #             data.select_dtypes(include=["object"]).columns
        #         ):
        #             try:
        #                 data[i] = pd.to_datetime(
        #                     data[i], infer_datetime_format=True, utc=False, errors="raise"
        #                 )
        #             except:
        #                 continue
        #         # drop datetime column
        #         data.drop(data.select_dtypes(include=["datetime64", "datetime64[ns, UTC]"]),axis=1,inplace=True)        
        # convert dtypes as of train data, le for target if classification prob,
        # for i in data.columns:
        #     data[i]=data[i].astype(self.final_dtypes[self.final_dtypes.index==i])
        
class Validation(BaseEstimator ):
    
    def __init__(self,target=None,ml_type='classification') -> None:
        self.ml_type=ml_type
        self.final_cols = []
        self.target = target
    def fit(self,dataset,y=None):
        logging.info("Preprocess step fit:Validation ")
        '''Performing filteration,drop target ->creates final col list'''
        
        data = dataset.drop(self.target,axis=1)
        data = data.loc[:,~data.columns.duplicated()]
        data.drop(data.select_dtypes(include=["datetime64", "datetime64[ns, UTC]"]),axis=1,inplace=True,errors='ignore')   
        self.final_training_columns = data.columns.to_list()
        

    def fit_transform(self, dataset, y=None):
        logging.info("Preprocess step fit_transform:Validation ")
        '''For fit_transform include target'''
        data = dataset
        target_col = data[self.target]
        targetless_cols = data.drop(self.target,axis=1)
        self.fit(data)
        df=  pd.concat((self.transform(targetless_cols),target_col),axis=1)
        return df 

    def transform(self,dataset,y=None):
        '''For testing data, do filteration'''
        logging.info("Preprocess step transform:Validation ")
        data = dataset
        for i in self.final_training_columns:
            if i not in data.columns:
                raise TypeError(
                    f"test data does not have column {i} which was used for training."
                )
        # from fit_tranform
        data = data[self.final_training_columns]

        data.replace([np.inf, -np.inf], np.NaN, inplace=True)

        # drop id column---
        return data
       
# add iterative and knn imputer
class Imputation(BaseEstimator ):
    
    """
    required:num_strat,cat_strat
    Imputes null values for numeric and categorical with sklearn's Simple Impute
    Num strats: mean,median,most_frequent,zero
    Cat strats: most_frequent,others
    """
    def __init__(
        self,
        target=None,
        num_strat='mean',
        cat_strat='most_frequent',
        num_fill=0,
        cat_fill="others",
    ):
        self.target=target,
        self.num_strat = num_strat
        self.cat_strat = cat_strat
        self.num_impute = SimpleImputer(
            strategy=self.num_strat,
            fill_value=num_fill,
        )
        self.cat_impute = SimpleImputer(
            strategy=cat_strat,
            fill_value=cat_fill,
        )
        self.num_fill = num_fill
        self.cat_fill=cat_fill

    def fit(self, dataset, y=None):
        logging.info("Preprocess step fit:Inputation")
        data = dataset
        self.numeric_columns = data.drop((self.target)[0], axis=1).select_dtypes(include=["number"]).columns
        self.categorical_columns = data.drop((self.target)[0], axis=1).select_dtypes(include=["object"]).columns

        if not self.numeric_columns.empty:
            self.num_impute.fit(data[self.numeric_columns])
        if not self.categorical_columns.empty:
            self.cat_impute.fit(data[self.categorical_columns])

    def transform(self, dataset, y=None):
        logging.info("Preprocess step transform:Inputation")
        data = dataset
        imputed_data = []
        if not self.numeric_columns.empty:
            numeric_data = pd.DataFrame(
                self.num_impute.transform(data[self.numeric_columns]),
                columns=self.numeric_columns,
                index=data.index,
            )
            imputed_data.append(numeric_data)
        if not self.categorical_columns.empty:
            categorical_data = pd.DataFrame(
                self.cat_impute.transform(data[self.categorical_columns]),
                columns=self.categorical_columns,
                index=data.index,
            )
            for col in categorical_data.columns:
                categorical_data[col] = categorical_data[col].apply(str)
            imputed_data.append(categorical_data)
        if imputed_data:
            data.update(pd.concat(imputed_data, axis=1))
        return data 

    def fit_transform(self,dataset,y=None):
        logging.info("Preprocess step fit_transform:Inputation")
        data = dataset
        target_col = data[(self.target)[0]]
        targetless_cols = dataset.drop((self.target)[0],axis=1)
        
        self.fit(data)
        df = pd.concat((self.transform(targetless_cols),target_col),axis=1)
        
        return df

class KnnImputation(BaseEstimator):
    pass

class IterativeImputation(BaseEstimator):
    pass

class ZeroVariance(BaseEstimator):
    def __init__(self,target=None) -> None:
        self.target=target
    def fit(self,dataset,y=None):
        '''For training data'''
        logging.info("Preprocess step fit:ZeroVariance")
        data = dataset.drop(self.target,axis=1)
        self.col_drop = []
        for i in data.select_dtypes(include=['number']).columns:
            if data[i].std()==0:
                self.col_drop.append(i)
       
    def transform(self,dataset,y=None):
        logging.info("Preprocess step transfomr:ZeroVariance")
        '''FOr testing data'''
        data= dataset.drop(self.col_drop,axis=1)
        return data
    def fit_transform(self,dataset,y=None):
        logging.info("Preprocess step fit_transform:ZeroVariance")
        data = dataset
        target_col = data[(self.target)]
        targetless_cols = data.drop(self.target,axis=1)
        self.fit(data)
        return pd.concat((self.transform(targetless_cols),target_col),axis=1)

class CatDummies(BaseEstimator):
    '''Cretes one hot encoding for categorical features '''
    def __init__(self,target=None):
        self.ohe = OneHotEncoder(dtype=np.float32,handle_unknown="ignore",)#
        self.target = target
        self.cols_to_drop = []
    def fit(self, dataset, y=None):      
        logging.info("Preprocess step fit:CatDummies")
        data = dataset.drop(self.target,axis=1)
        # will only do this if there are categorical variables
        if len(data.select_dtypes(include=("object")).columns) > 0:
            for i in data.select_dtypes(include=("object")).columns:
                # check for unique values
                if data[i].nunique() >= data.shape[0]:
                    self.cols_to_drop.append(i)
                    data = data.drop(i,axis=1,errors='ignore')
            self.data_nonc = data.select_dtypes(exclude=("object"))
            categorical_data = data.select_dtypes(include=("object"))
            # # now fit the trainin column
            self.ohe.fit(categorical_data)
            self.data_columns = self.ohe.get_feature_names(categorical_data.columns)

    def transform(self, dataset, y=None):
        logging.info("Preprocess step transform:CatDummies")
        data = dataset
        # will only do this if there are categorical variables
        if len(data.select_dtypes(include=("object")).columns) > 0:
            if len(self.cols_to_drop) >=1:
                data = data.drop(self.cols_to_drop,axis=1,errors='ignore')
            # only for test data
            self.data_nonc = data.select_dtypes(exclude=("object"))
            # fit only categorical columns
            array = self.ohe.transform(
                data.select_dtypes(
                    include=("object")
                )
            ).toarray()
            data_dummies = pd.DataFrame(array, columns=self.data_columns)
            data_dummies.index = self.data_nonc.index
            
            # now put target , numerical and categorical variables back togather
            data = pd.concat(( self.data_nonc, data_dummies), axis=1)
            del self.data_nonc
            return data
        else:
            return data
    def fit_transform(self, dataset, y=None):
        logging.info("Preprocess step fit_transform:CatDummies")
        data = dataset
        target_col = data[self.target]
        targetless_cols = data.drop(self.target,axis=1)
        self.fit(data)
        return pd.concat((self.transform(targetless_cols),target_col),axis=1)

class ScaleTransformer(BaseEstimator):
    '''Applies, minmax,zscore,yeo-johnson, quantile'''
    def __init__(self,target=None,function ='zscore') -> None:
        self.function = function
        self.target = target
        
    def fit(self,dataset,y=None):
        '''For training data, apply transformation'''
        logging.info("Preprocess step fit:ScaleTransformer")
        data = dataset.drop(self.target,axis=1)
        self.num_cols = data.select_dtypes(include=['number']).columns
        
        if len(self.num_cols)>0:
          
          if self.function=='zscore':
            self.scaler = StandardScaler()
          elif self.function =='minmax':
            self.scaler = MinMaxScaler()
          elif self.function =='yeo-johnson':
            self.scaler = PowerTransformer(method="yeo-johnson", standardize=True)
          elif self.function =='quantile':
            self.scaler =QuantileTransformer(output_distribution="normal",random_state=45)
        self.scaler.fit(data[self.num_cols])
        
    
    def transform(self,dataset,y=None):
        '''FOr testing data, apply transformer used in training'''
        logging.info("Preprocess step transform:ScaleTransformer")

        data = dataset
        # len of num cols is >0
        if len(self.num_cols)>0:
            num_data = pd.DataFrame(self.scaler.transform(data[self.num_cols]))
            num_data.index = data.index
            num_data.columns = self.num_cols
            for i in self.num_cols:
                data[i] = num_data[i]
        
        return data
    
    def fit_transform(self,dataset,y=None):
        logging.info("Preprocess step fit_transform:ScaleTransformer")

        data = dataset
        
        target_col = data[self.target]
        
        targetless_cols = data.drop(self.target,axis=1)
        self.fit(data)
        c = pd.concat((self.transform(targetless_cols),target_col),axis=1)
        
        return c

class OutlierRemoval(BaseEstimator):
    pass

class PerfectCollinearity(BaseEstimator):
    '''Removes first column which have perfect i.e, 1 collinearity between them(features)
       One hot encoded variables only
    '''
    def __init__(self,target=None) -> None:
        self.col_drop = []
        self.target = target
    def fit(self,dataset,y=None):
        logging.info("Preprocess step fit:PerfectCollinearity")

        data = dataset.drop(self.target,axis=1)
        if len(data.columns) <= 2:
            return data
        corr = pd.DataFrame(np.corrcoef(data.T))
        corr.columns = data.columns
        corr.index = data.columns
        corr_matrix = abs(corr)

        # Now, add a column for variable name and drop index
        corr_matrix["column"] = corr_matrix.index
        corr_matrix.reset_index(drop=True, inplace=True)

        # now we need to melt it , so that we can correlation pair wise , with two columns
        cols = corr_matrix.column
        melt = corr_matrix.melt(id_vars=["column"], value_vars=cols).sort_values(
            by="value", ascending=False
        )  # .dropna()
        melt["value"] = round(melt["value"], 2)  # round it to two digits

        # now pick variables where value is one and 'column' != variabe ( both columns are not same)
        c1 = melt["value"] == 1.00
        c2 = melt["column"] != melt["variable"]
        melt = melt[((c1 == True) & (c2 == True))]

        # we need to now eleminate all the pairs that are actually duplicate e.g cor(x,y) = cor(y,x) , they are the same , we need to find these and drop them
        melt["all_columns"] = melt["column"] + melt["variable"]

        # this puts all the coresponding pairs of features togather , so that we can only take one, since they are just the duplicates
        melt["all_columns"] = [sorted(i) for i in melt["all_columns"]]

        # # now sort by new column
        melt = melt.sort_values(by="all_columns")

        # # take every second colums
        melt = melt.iloc[::2, :]

        # lets keep the columns on the left hand side of the table
        self.col_drop = melt["variable"]

    def transform(self,dataset,y=None):
        logging.info("Preprocess step transform:PerfectCollinearity")

        return dataset.drop(self.col_drop,axis=1)
    def fit_transform(self, dataset, y=None):
        logging.info("Preprocess step fit_transform:PerfectCollinearity")

        data = dataset
        target_col = data[self.target]
        targetless_cols = data.drop(self.target,axis=1)
        self.fit(data)
        return pd.concat((self.transform(targetless_cols),target_col),axis=1)


class ClusterData(BaseEstimator):
    def __init__(self,target=None) -> None:
        '''Create cluster for data (min:2,max:20)'''
        self.target = target
        # self.n_clusters = 3 #change to sqrt 
    def fit(self,dataset,y=None):
        logging.info("Preprocess step fit:ClusterData")
        wcss = []
        data = dataset.drop(self.target,axis=1,errors='ignore')
        print("-----------------Fit cluster data:",data.shape)
        # for i in range (2,10):  #hard coded
        #     kmeans=KMeans(n_clusters=i,init='k-means++',random_state=42) # initializing the KMeans object
        #     kmeans.fit(data) # fitting the data to the KMeans Algorithm
        #     wcss.append(kmeans.inertia_)
        # self.kn = KneeLocator(range(1, 10), wcss, curve='convex', direction='decreasing')
        self.kmeans = KMeans(n_clusters=5,init='k-means++',random_state=42)
        self.kmeans.fit(data)

    def transform(self,dataset,y=None):
        logging.info("Preprocess step transform:ClusterData")
        data = dataset
        data['cluster_label']=self.kmeans.predict(data)
        return data
    def fit_transform(self,dataset,y=None):
        logging.info("Preprocess step fit_transfomrm:ClusterData")
        data = dataset
        target_col = data[(self.target)]
        targetless_cols = data.drop(self.target,axis=1)
        self.fit(data)
        return pd.concat((self.transform(targetless_cols),target_col),axis=1)

class FeatureSelection(BaseEstimator):
    pass

class Multicollinearity(BaseEstimator):
    pass

class ImbalancedData(BaseEstimator):
    pass
    # over = SMOTE(sampling_strategy=0.1)
    # under = RandomUnderSampler(sampling_strategy=0.5)
    # steps = [('o', over), ('u', under)]
    # pipeline = Pipeline(steps=steps)

    # # transform the dataset
    # X, y = pipeline.fit_resample(data.drop(self.target,axis=1),self.target)

class TargetTreat(BaseEstimator):
    def __init__(self,target=None,ml_type='classification') -> None:
        self.target = target
        self.ml_type = ml_type
        self.le = LabelEncoder()
    def fit(self,dataset,y=None):
        logging.info("Preprocess step fit:TargetTreat")
        data = dataset
        if ((self.target !=None) and (data[(self.target)].dtypes=='object')):
        # remove row where target is NaN
            try:
                data.dropna(subset=[self.target], inplace=True)
                logging.warning("Found nan rows in target")
            except KeyError:
                pass
            self.le.fit(data[self.target])
            self.replacement = _get_labelencoder_reverse_dict(self.le)
    def transform(self,dataset,y=None):
        logging.info("Preprocess step transform:TargetTreat")
        data = dataset
        if (self.target) in data.columns:
            try:
                data.dropna(subset=[self.target], inplace=True)
            except KeyError:
                pass
            if (data[self.target].dtypes=='object'):
                data[self.target] = self.le.transform(data[self.target])      
        return data
    def fit_transform(self,dataset,y=None):
        logging.info("Preprocess step fit_transform:TargetTreat")
        self.fit(dataset)                
        return self.transform(dataset)

def _get_labelencoder_reverse_dict(le: LabelEncoder) -> dict:
    # now get the replacement dict
    rev = le.inverse_transform(range(0, len(le.classes_)))
    rep = np.array(range(0, len(le.classes_)))
    replacement = {}
    for i, k in zip(rev, rep):
        replacement[i] = k
    return replacement

def preprocess_path(ml_type,
scale_data=True,
scaling_method="zscore",
numeric_imputation_strategy="mean",
dummify_categoricals=True,
remove_perfect_collinearity=False,
target=None,data=None,
cluster_entire_data=False,zerovar=True):
   
    validates = Validation(
        ml_type=ml_type,
        target=target,
    )
    imputer = Imputation(
        target=target
    )
    if scale_data == True:
        scaling = ScaleTransformer(
            target=target,
            function=scaling_method,
        )
    else:
        scaling = SKLEARN_EMPTY_STEP
    
    if dummify_categoricals == True:
        dummy = CatDummies(target =target)
    else:
        dummy = SKLEARN_EMPTY_STEP

    if cluster_entire_data == True:
        cluster_all = ClusterData(
            target = target
        )
    else:
        cluster_all = SKLEARN_EMPTY_STEP
    if remove_perfect_collinearity == True:
        fix_perfect = PerfectCollinearity(target=target)
    else:
        fix_perfect = SKLEARN_EMPTY_STEP
    if zerovar==True:
        zerovar = ZeroVariance(target=target)
    else:
        zerovar = SKLEARN_EMPTY_STEP
    target_treat = TargetTreat(target=target,ml_type=ml_type)
    pipe = Pipeline(
         [   
            ("validates", validates),
            ("zerovar",zerovar),
            ("imputer", imputer),
            ("dummy", dummy),
            ("scaling", scaling),
            ("fix_perfect", fix_perfect),
            
            ("cluster_all", cluster_all),
            ("target_treat",target_treat),
            
        ]
    )    
    return pipe

